/* Functions to compute midpoint edginess with minimal (4-point) mask.

   Derived from the article
   "Singularity Detection in Images Using Dual Local Autocovariance"
   by  Wojciech Czaja and Mladen Victor Wickerhauser.

   Programmer: M.V.Wickerhauser <victor@math.wustl.edu>
   Date: 22 December 2002

 */

#include <assert.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>

#define WAY_TOO_BIG	4097	/* Too many rows or columns */

typedef float real;

struct Grid {
  real **samples;		/* indexed `samples[r][c]' */
  int rows;			/* number of rows = 1+r1-r0 */
  int cols;			/* number of columns = 1+c1-c0 */
  int r0;			/* least valid `r' index */
  int r1;			/* greatest valid `r' index */
  int c0;			/* least valid `c' index */
  int c1;			/* greatest valid `c' index */
};
typedef	struct Grid	*Grid;

Grid	newGrid		( int, int, int, int, const real * );
void	freeGrid	( Grid );
void	edgify		( Grid );
real	anormalize	( Grid, int );
Grid	readPGM		( FILE * );
void	writePGM	( FILE *, Grid );


/* Allocate a Grid data structure and its samples array. */
Grid
newGrid ( int r0, int r1, int c0, int c1, /* Row and column limits */
	  const real *in)	/* if non-NULL, copy these values */
{
  Grid new;
  int k;

  assert(r0<=r1);  assert(c0<=c1);
  new = (Grid)malloc(sizeof(struct Grid)); assert(new);
  new->r0 = r0;   new->r1=r1;   new->c0=c0;   new->c1=c1;
  new->rows = 1 + r1 - r0;      new->cols = 1 + c1 - c0;
  new->samples = (real **) malloc ( new->rows * sizeof(real *) );
  assert(new->samples);
  for(k=0; k<new->rows; k++) {
    new->samples[k] = (real *) calloc ( new->cols, sizeof(real) );
    assert(new->samples[k]);
    new->samples[k] -= c0;	/* `samples[k][c]' now valid for c0<=c<=c1 */
  }
  new->samples -= r0;		/* `samples[r][*]' now valid for r0<=r<=r1 */

  if(in) {			/* assume the data is stored row-by-row */
    int i,j;
    for(i=r0;i<=r1; i++)
      for(j=c0; j<=c1; j++)
	new->samples[i][j] = *in++;
  } /* ... otherwise, leave `new->samples[][]' filled with zeroes. */
  return new;
}

/* Deallocate a Grid data structure and its samples array. */
void
freeGrid ( Grid in )
{
  int k;
  if(in) {
    for(k=in->r0; k<=in->r1; k++ )
      free( in->samples[k] + in->c0 );
    free( in->samples + in->r0 );
    free(in);
  }
  return;
}

    
/* Read a PGM file into a Grid data structure, with no error checking.
   The file's header must be exactly
	P5
	<width> <height>
	255
	<byte-data>
*/
Grid
readPGM ( FILE *pgmfile )
{
  int i,j, cols,rows;
  Grid out;
  unsigned char *buf, *bp;

  assert(pgmfile);

  fscanf( pgmfile, "P5\n%d%d\n255\n", &cols,&rows );
  assert(cols>0);  assert(cols<WAY_TOO_BIG);
  assert(rows>0);  assert(rows<WAY_TOO_BIG);

  bp = buf = (unsigned char *)malloc(rows*cols);  assert(buf);
  fread(buf, 1, rows*cols, pgmfile);

  out = newGrid( 0, rows-1, 0, cols-1, 0 );
  
  for(i=0; i<rows; i++)
    for(j=0; j<cols; j++)
      out->samples[i][j] = (real) *bp++;

  free(buf);
  return out;
}

/* Write a PGM file from data in a Grid data structure,
   assumed normalized. The file's header will be exactly
	P5
	<cols> <rows>
	255
	<byte-data>
*/
void
writePGM ( FILE *pgmfile, Grid in )
{
  int i,j;
  assert(pgmfile);

  fprintf(pgmfile,"P5\n%d %d\n255\n", in->cols,in->rows); /* Header */
  for(i=in->r0; i<=in->r1; i++)
    for(j=in->c0; j<=in->c1; j++)
      putc((int)(0.5+in->samples[i][j]), pgmfile); /* Byte-data */

  return;
}

/* Affine-normalize the sample values in a Grid to fill the dynamic range of
   8-bit grayscale display devices.  Assume that all values are nonnegative.
   Use linear scaling with an offset.  If reverse==0, map the smallest value
   to 0 and the largest value to 255.  Otherwise, reverse the mapping so that
   the smallest value becomes 255 and the largest value becomes 0.  Return
   the original difference between the largest and smallest values. */
real
anormalize(Grid g, int reverse)
{
  int i,j;
  real x, xmax, xmin, norm;

  assert(g);  assert(g->samples);

  /* Find maximum and minimum */
  xmax = xmin = g->samples[g->r0][g->c0];
  for ( i = g->r0  ;  i <= g->r1  ;  i++ )
    for ( j = g->c0  ;  j <= g->c1  ;  j++ ) {
      x = g->samples[i][j];
      if( x>xmax ) xmax = x;
      else
	if( x<xmin ) xmin = x;
    }
  assert ( xmin >= 0 );		/* Are all sample values nonnegative? */
  if ( xmin > 0 ) {		/* Else no offset is needed */
    for ( i = g->r0  ;  i <= g->r1  ;  i++ )
      for ( j = g->c0  ;  j <= g->c1  ;  j++ )
	g->samples[i][j] -= xmin;
    xmax -= xmin;
  }
  if ( xmax > 0 ) {		/* Else grid is all zeroes; do not rescale */
    norm = 255.4/xmax;
    for ( i = g->r0  ;  i <= g->r1  ;  i++ )
      for ( j = g->c0  ;  j <= g->c1  ;  j++ )
	g->samples[i][j] *= norm;
  }
  if(reverse) {			/* Else do not reverse the values  */
    for ( i = g->r0  ;  i <= g->r1  ;  i++ )
      for ( j = g->c0  ;  j <= g->c1  ;  j++ )
	g->samples[i][j] = 255.4 - g->samples[i][j];
  }
  return xmax;
}


/* Return the difference between the eigenvalues of the dual local
   autocovariance matrix computed for a 4-point subgrid in an image,
   symmetric about a midpoint. */
static real
evdiff( real f[2][2] )
{
  real determinant, crossdiff;
  determinant = f[0][0]*f[1][1] - f[0][1]*f[1][0];
  crossdiff = (f[0][0]-f[1][1])*(f[0][1]-f[1][0]);
  return sqrt(determinant*determinant + crossdiff*crossdiff);
}

/* Replace the image in a grid with the midpoint edginess function.  Treat
   the right-hand and bottom boundaries by midpoint symmetric extension.  */
void
edgify( Grid g )
{
  int n,m, i, j;
  real f[2][2];

  assert(g);

  /* Main part -- all but last row: */
  for ( m = g->r0  ;  m < g->r1  ;  m++ ) {
    /* Main part -- all but last column: */
    for ( n = g->c0  ;  n < g->c1  ;   n++ ) {
      for(i=0; i<2; i++)
	for(j=0; j<2; j++)
	  f[i][j] = g->samples[m+i][n+j];
      g->samples[m][n] = evdiff(f);
    }
    /* Right boundary -- last column of all but last row: */
    f[0][0] = f[0][1] = g->samples[m][n];
    f[1][0] = f[1][1] = g->samples[m+1][n];
    g->samples[m][n] = evdiff(f);
  }
  /* Bottom boundary -- last row of all but last column: */
  for ( n = g->c0  ;  n < g->c1  ;   n++ ) {
    f[0][0] = f[1][0] = g->samples[m][n];
    f[0][1] = f[1][1] = g->samples[m][n+1];
    g->samples[m][n] = evdiff(f);
  }
  /* Bottom right-hand corner -- last row, last column: */
  g->samples[m][n] = 0;

  return;
}



/* Read an image in PGM format from stdin.
   Compute the edginess and output it to stdout, in PGM format.
 */
int
main(void)
{
  Grid image;

  image = readPGM( stdin );
  edgify( image );
  anormalize ( image, 1 );
  writePGM ( stdout, image);
  freeGrid(image);

  return 0;
}
